using System;
using System.Diagnostics;
using System.Text;
using System.Web;
using System.Web.Caching;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using System.Xml;
using System.Configuration;
using System.IO;
using System.Reflection;
using System.ComponentModel;

namespace XmlTemplate.WebControls {
	public class BasePage : System.Web.UI.Page {
		private string				_PageTitle;
		private string				_PageMetaData;
		private string				_PageStyleSheet;
		private string				_PageJavaScript;
		private string				_PageBodyAttributes;
		private BaseHtml			_Html;
		private string				_XmlTemplate = null;
		private ValidatorCollection	_validators = null;
		private ControlCollection	_Controls = null;

		#region Constructor
		public BasePage() {
			HttpContext context = HttpContext.Current;
			if (context != null) { //For designer
				_Controls = new ControlCollection(this);

				//QueryString Xml template takes precedence over ones stored in cookie
				if (context.Request.QueryString["XmlTemplate"] != null) {
					this.XmlTemplate = context.Server.MapPath(context.Request.QueryString["XmlTemplate"].ToString());
				} else if (context.Request.Cookies["XmlTemplate"] != null) {
					this.XmlTemplate = context.Server.MapPath(context.Request.Cookies["XmlTemplate"].Value);
				}
				//If we have no XmlTemplate path get it from web.config
				if (this.XmlTemplate == null) {
					string template = ConfigurationSettings.AppSettings["XmlTemplate"];
					if (template != null) this.XmlTemplate = context.Server.MapPath(template);
				}
			}
		}
		#endregion

		#region  Properties
		public BaseHtml Html {
			get {
				return _Html;
			}
			set {
				_Html = value;
			}
		}

		public string XmlTemplate {
			get {
				return _XmlTemplate;
			}
			set {
				_XmlTemplate = value;
			}
		}

		public string PageTitle {
			get {
				return _PageTitle;
			}
			set {
				_PageTitle = value;
			}
		}

		public string PageMetaData {
			get {
				return _PageMetaData;
			}
			set {
				_PageMetaData = value;
			}
		}

		public string PageStyleSheet {
			get {
				return _PageStyleSheet;
			}
			set {
				_PageStyleSheet = value;
			}
		}

		public string PageJavaScript {
			get {
				return _PageJavaScript;
			}
			set {
				_PageJavaScript = value;
			}
		}

		public string PageBodyAttributes {
			get {
				return _PageBodyAttributes;
			}
			set {
				string atts  = value.Replace("\"", "");
				atts = atts.Replace("'", "");
				this._PageBodyAttributes = atts;
			}
		}
		# endregion

		# region Overrides

		protected override void AddParsedSubObject(Object Control) {
			if (this.XmlTemplate != null) {
				//Store page controls in special collection so that their Init() method isn't
				//called twice (once when added to page control collection and again when 
				//they are added into the BasePage control collection)
				_Controls.Add((System.Web.UI.Control)Control);
			} else {
				//If no template is specified don't break the process
				//Instead, add controls into base Page control collection and
				//continue with normal processing.
				base.Controls.Add((System.Web.UI.Control)Control);
			}
		}

		protected override void OnInit(EventArgs e) {
			base.OnInit(e);

			if (this.XmlTemplate != null) {
				//Handle grabbing Page object validators
				if (this.Validators.Count > 0) {
					_validators = new ValidatorCollection();
					foreach (BaseValidator validator in this.Validators) {
						_validators.Add(validator);
					}
				}

				this.Html = ParseXmlTemplate();

				//Handle re-adding validators to BasePage
				if (_validators != null) {
					foreach (BaseValidator validator in _validators) {
						this.Validators.Add(validator);
					}
				}

				this.Controls.Add(this.Html);
			}
		}

		protected override void OnPreRender(EventArgs e) {
			base.OnPreRender(e);

			if (this.XmlTemplate != null) {
				if (this.PageTitle != null) {
					Html.Head.Title.Controls.Clear();
					Html.Head.Title.Controls.Add(new LiteralControl(this.PageTitle));
				}

				if (this.PageJavaScript != null) {
					StringBuilder sbJS = new StringBuilder();
					string[] jsPaths = _PageJavaScript.Split( new char[]{','});
					foreach (string jsPath in jsPaths) {
						sbJS.Append("<script language=\"javascript\" src=\"");
						sbJS.Append(jsPath);
						sbJS.Append("\"></script>");
						this.Html.Head.Custom.Controls.Add(new LiteralControl(sbJS.ToString()));																		  
					}
				}

				if (this.PageStyleSheet != null) {
					StringBuilder sbCSS = new StringBuilder();
					string[] cssPaths = this.PageStyleSheet.Split(new char[]{','});
					foreach (string cssPath in cssPaths) {
						sbCSS.Append("<link rel=\"stylesheet\" type=\"text/css\" href=\"");
						sbCSS.Append(cssPath);
						sbCSS.Append("\" />");
						this.Html.Head.Custom.Controls.Add(new LiteralControl(sbCSS.ToString()));
					}
				}

				if (this.PageBodyAttributes != null) {
					try {
						string[] atts = this.PageBodyAttributes.Split(new char[]{' '});
						foreach (string att in atts) {
							string[] nameValue = att.Split(new char[]{'='});
							string name = nameValue[0];
							string val = nameValue[1];
							Html.Body.Attributes.Add(name, val);
						}
					}
					catch {}
				}
			}
		}
		# endregion

		#region Parsing

		private Stream CopyStream(MemoryStream ms) {
			MemoryStream msTemp = new MemoryStream();
			ms.WriteTo(msTemp);
			msTemp.Position = 0;
			return msTemp;
		}

		private Stream GetXmlTemplate() {
			if (this.Cache.Get(this.XmlTemplate) != null) {
				MemoryStream ms = (MemoryStream)this.Cache.Get(this.XmlTemplate);
				ms.Position = 0;
				return CopyStream(ms);
			} else {
				FileStream fs = null;
				try {
					fs = new FileStream(this.XmlTemplate,FileMode.Open,FileAccess.Read);
					MemoryStream ms = new MemoryStream();
					byte[] bytes = new byte[1024];
					int i;
					while ((i = fs.Read(bytes,0,bytes.Length)) != 0) {
						ms.Write(bytes,0,i);
					}
					ms.Position = 0;
					CacheDependency dep = new CacheDependency(new string[]{this.XmlTemplate});
					this.Cache.Insert(this.XmlTemplate,ms,dep);
					return CopyStream(ms);
				}
				catch (Exception exp) {
					throw new ApplicationException(String.Concat("Unable to cache XML template (",exp.Message,"): ",this.XmlTemplate));
				}
				finally {
					fs.Close();
				}
			}
		}

		private BaseHtml ParseXmlTemplate() {
			Trace.Warn("Start ParseXmlTemplate");
			bool foundContent = false;
			XmlTextReader reader = null;
			BaseHtml root = null;	
			Control currentControl = null;
			try {
				reader = new XmlTextReader(GetXmlTemplate());
				object uriText = reader.NameTable.Add("http://www.xmlforasp.net/templates");
				object ContentPlaceHolderText = reader.NameTable.Add("contentplaceholder");
				object PlaceHolderText = reader.NameTable.Add("placeholder");
				object UserControlText = reader.NameTable.Add("usercontrol");
				object ServerControlText = reader.NameTable.Add("servercontrol");
				object TypeAttributeText = reader.NameTable.Add("type");
				object htmlText = reader.NameTable.Add("html");
				object headText = reader.NameTable.Add("head");
				object titleText = reader.NameTable.Add("title");
				object bodyText = reader.NameTable.Add("body");
				object formText = reader.NameTable.Add("form");
				while (reader.Read()) {
					if (reader.NodeType != XmlNodeType.Whitespace && reader.NodeType != XmlNodeType.Comment) {
						if (reader.NamespaceURI.ToLower().Equals(uriText)) {
							string id = reader.GetAttribute("id");
							if (reader.LocalName.ToLower().Equals(ContentPlaceHolderText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									PlaceHolder ph = new PlaceHolder();
									if (id != null) ph.ID = id;
									foundContent = true;
									if (root.Body.Form == null) root.Body.Form = new BaseForm();
									root.Body.Form.Content = ph;
									AddControls(ph);
								} else if (reader.NodeType == XmlNodeType.EndElement) {
									currentControl = currentControl.Parent;
								}
							} else if (reader.LocalName.ToLower().Equals(UserControlText)) {
								try {
									Control uc = LoadControl(reader.GetAttribute("src"));
									if (id != String.Empty) {
										uc.ID = id;
									}
									currentControl.Controls.Add(uc);
								}
								catch (Exception exp) {
									throw new ApplicationException("Error loading template user control.", exp);
								}
							} else if (reader.LocalName.ToLower().Equals(ServerControlText)) {
								string typeVal = reader.GetAttribute("type");
								try {
									Type type = null;
									object serverCtl = null;
									if (this.Cache.Get(typeVal) == null) {
										string typeInfo = typeVal.Substring(0,typeVal.IndexOf(",")).Trim();
										string assemblyInfo = typeVal.Substring(typeVal.IndexOf(",")+1).Trim();
										Assembly assembly = LoadAssembly(assemblyInfo);
										type = assembly.GetType(typeInfo,true,true);
										ConstructorInfo ci = type.GetConstructor(new Type[0]);
										serverCtl = ci.Invoke(null);
										this.Cache.Insert(typeVal,ci); 
									} else {
										serverCtl = ((ConstructorInfo)this.Cache[typeVal]).Invoke(null);
										type = serverCtl.GetType();
									}
									while (reader.MoveToNextAttribute()) {
										if (!reader.LocalName.ToLower().Equals(TypeAttributeText)) { //Don't grab type attribute
											PropertyInfo propInfo = type.GetProperty(reader.LocalName,
												BindingFlags.Public | BindingFlags.Instance | BindingFlags.IgnoreCase);
											if (propInfo != null) {
												ParameterInfo[] pi = propInfo.DeclaringType.GetProperty(reader.LocalName, 
													BindingFlags.Public | BindingFlags.Instance | BindingFlags.IgnoreCase).GetSetMethod().GetParameters();
												object o = TypeDescriptor.GetConverter(pi[0].ParameterType).ConvertFrom(reader.Value);
												propInfo.SetValue(serverCtl,o,null);
											}
										}
									}
									currentControl.Controls.Add((Control)serverCtl);
								}
								catch (Exception exp) {
									throw new ApplicationException("Error loading template server control: " + typeVal, exp);
								}
							}
						}
						else {
							if (reader.LocalName.ToLower().Equals(htmlText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									root = new BaseHtml();
									currentControl = root;
								}
							} else if (reader.LocalName.ToLower().Equals(headText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									BaseHead baseHead = new BaseHead();
									baseHead.Custom = new PlaceHolder();
									root.Head = baseHead;
									currentControl = baseHead;
								}
							} else if (reader.LocalName.ToLower().Equals(titleText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									HtmlGenericControl title = new HtmlGenericControl("title");
									root.Head.Title = title;
									currentControl = title;
								}
								else {
									currentControl = root.Head;
								}
							} else if (reader.LocalName.ToLower().Equals(bodyText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									BaseBody baseBody = new BaseBody();
									while (reader.MoveToNextAttribute()) {
										baseBody.Attributes.Add(reader.Name, reader.Value);
									}
									reader.MoveToElement();
									root.Body = baseBody;
									currentControl = baseBody;
									//If we don't have a head go ahead and add it
									if (root.Head == null) {
										root.Head = new BaseHead();
										root.Head.Title = new HtmlGenericControl("title");
									}
								}
							} else if (reader.LocalName.ToLower().Equals(formText)) {
								if (reader.NodeType == XmlNodeType.Element) {
									string formID = reader.GetAttribute("id");
									BaseForm baseForm = new BaseForm();
									baseForm.ID = (formID == String.Empty)?"pageForm":formID;
									root.Body.Form = baseForm;
									currentControl = baseForm;
								} else if (reader.NodeType == XmlNodeType.EndElement) {
									currentControl = currentControl.Parent;
								}
							} else {
								if (currentControl != null) {
									LiteralControl litControl = new LiteralControl();
									if (reader.NodeType == XmlNodeType.Element) {
										litControl.Text = WriteStartElement(reader);
									}
									else if (reader.NodeType == XmlNodeType.Text) {
										litControl.Text = reader.Value;
									}
									else if (reader.NodeType == XmlNodeType.EndElement) {
										litControl.Text = WriteEndElement(reader.LocalName);
									}
									currentControl.Controls.Add(litControl);
								}
							}
						}
					}
				} //while
			} //try
			catch (Exception exp) {
				throw new ApplicationException("Template file cannot be parsed:" + exp.Message);
			}
			finally {
				if (reader != null) {
					reader.Close();
				}
			}
			Trace.Warn("End ParseXmlTemplate");

			if (foundContent == false) {
				throw new ApplicationException("Template Error: You must add a ContentPlaceHolder tag with an id of \"MainContent\" into the template.");
			}
			else {
				return root;
			}
		}

		private Assembly LoadAssembly(string assemblyInfo) {
			try {
				return Assembly.Load(assemblyInfo);
			}
			catch {
				return Assembly.LoadWithPartialName(assemblyInfo);
			}
		}

		private string WriteStartElement(XmlReader reader) {
			if (reader.IsEmptyElement) {
				return reader.ReadOuterXml();
			}
			else {
				StringWriter sw = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(sw);
				writer.WriteStartElement(reader.LocalName);
				writer.WriteAttributes(reader,false);
				writer.WriteEndElement();
				writer.Close();
				return sw.ToString();
			}
		}

		private string WriteEndElement(string elementName) {
			return String.Concat("</", elementName, ">");
		}

		private void AddControls(PlaceHolder content) {
			Trace.Warn("Start Add Controls");
			int count = _Controls.Count;
			int i  = -count;
			while (i < count) {
				Control currentControl = _Controls[0];
				if (currentControl is HtmlForm) {
					int childCount = currentControl.Controls.Count;
					int j = -childCount;
					while (j < childCount) {
						Control child = currentControl.Controls[0];
						AddControl(child,content);
						currentControl.Controls.Remove(child);
						childCount = currentControl.Controls.Count;
						j++;
					}
					//Fallback
				} else {
					AddControl(currentControl,content);
				}
				_Controls.Remove(currentControl);
				count = _Controls.Count;
				i++;
			}
			Trace.Warn("End Add Controls");
		}

		private void AddControl(Control ctl,PlaceHolder content) {
			//Handle cached user controls (when tracing turned on) as they cause 
			//duplicate IDs when control itself is added into control collection			
			/*if (ctl is StaticPartialCachingControl) {
				//StaticPartialCachingControl contains main control as a child
				if (ctl.Controls != null && ctl.Controls.Count > 0) {
					content.Controls.Add(ctl.Controls[0]);
				}
			} else */
			if (ctl is PartialCachingControl) {
				if (((PartialCachingControl)ctl).CachedControl != null) {
					content.Controls.Add(((PartialCachingControl)ctl).CachedControl);
				}
			} else {
				content.Controls.Add(ctl);
			}
		}


		#endregion
	}
}
